/*
	Epocrates Mobile JavaScript API
*/

EPOC = {
	UI: {},
	Util: {},
	Dom: {},
	db: {}
};

// ultra simple subclassing sugar
Function.prototype.extend = function(superclass, proto) {
	this.prototype = new superclass();

	// optional subclass methods and properties
	if (proto)
		for (var i in proto)
			this.prototype[i] = proto[i];
};

/**
	Global constants used in Core, required along with epoc.js for most everything.

	@class EPOC.Constants
	@static
*/
EPOC.Constants = {

	/** @property touchDomEvent */
	touchDomEvent: "click",
	
	/** @property selectDomEvent */
	selectDomEvent: "change",

	/** @property holdDomEvent */
	holdDomEvent: "mousedown",

	/** @property flickStartDomEvent */
	flickStartDomEvent: "mousedown",

	/** @property moveDomEvent */
	moveDomEvent: "mousemove",

	/** @property blurDomEvent */
	blurDomEvent: "blur",

	/** @property changeDomEvent */
	changeDomEvent: "change",

	/** @property focusDomEvent */
	focusDomEvent: "focus",
	
	/** @property typeDomEvent */
	typeDomEvent: "keyup",

	/** @property typeDomEvent */
	errorDomEvent: "error",

	/** @property downDomEvent */
	downDomEvent: "mousedown",

	/** @property upDomEvent */
	upDomEvent: "mouseup",
	
	/** @property maximumStackDepth */
	maximumStackDepth: 20
};/**
	Event handling functions, either DOM or custom. Can optionally pass
	static data, and resolve object context when called.
	
	@namespace EPOC
	@module Event
	@copyright 2009 Epocrates, Inc.
*/

/**
	@class Event
	@static
*/
EPOC.Event = {
	/**
		Sort out the target of a given DOM event.
		
		@method getTarget
		@param {DOM Event} e
		@return {DOM Element}
	*/
	getTarget: function(e) {
		return e.target ? e.target : e.srcElement;
	},

	/**
		Setup a new DOM event listener for a given element.
		
		@method on
		@param {DOM Element} element
		@param {string} event E.g. mousedown, click, change, blur
		@param {Function} call
		@param {Object} context
		@param {any} data
	*/
	on: function(element, event, call, context, data) {
		if (!call || !element)
			return;

		var element = EPOC.Dom.get(element);
		var call = call;
		var data = data || "";

		var wrappercall = function(e) {
			if (context)
				call.call(context, e, data);
			else
				call(e, data);
		};
		
		if (element.addEventListener)
			element.addEventListener(event, wrappercall, true);
		else
			element.attachEvent("on" + event, wrappercall);
	},
	
	/**
		Stop event propogation, used inside a DOM event handler.

		@param {DOM Event} e
	*/
	stop: function(e) {
		if (e.stopPropagation)
			e.stopPropagation();
		else
			e.cancelBubble = true;
	},
	
	/**
		Make a call asynchonously using window.setTimeout().
		
		@param {Function} call
		@param {Object} context
		@param {integer} delay in milliseconds (100ms is the default)
	*/
	yield: function(call, context, delay) {
		if (!delay)
			var delay = 100;
		if (!context)
			var context = this;
			
		var timer = window.setTimeout(function() {
			call.call(context);
		}, delay);
		
		return timer;
	}
}


/**
	Allows an object to create an event which observers can subscribe to. The most common event
	is "changeEvent". The subject is not directly aware of its observers (nor does it need to be),
	this class takes care of all that.
	
	Note that this functionality is not coupled with any particular data class, but remains loose
	so that it can be more flexible. Typically, a subject will consume one or more Event.Subject
	objects for each event exposed for subscription, see <pre>EPOC.UI.ListView</pre>.
		
	@copyright 2009 Epocrates, Inc.
	@namespace EPOC.Event
	@class Subject
	@constructor
	@param {Object} subject The object that will be firing off event notifications.
*/
EPOC.Event.Subject = function(subject) {
	this.subscriptions = [];
	this.subject = subject;	
};

EPOC.Event.Subject.prototype = {
	/**
		Start sending events to an observer.

		@method subscribe
		@param {Function} call
		@param {Object} observer
		@param {any} data Optional data to be passed to the observer when this event is fired.
		@return {boolean} Success is true.
	*/
	subscribe: function(call, observer, data) {
		if (!call)
			return false;
		
		var observer = observer || this;
		if (typeof data == 'undefined')
			var data = "";
		
		this.subscriptions.push({ "call": call, "observer": observer, "data": data});
	
		return true;
	},
	
	/**
		Stop sending event notifications to an observer.

		@method unsubscribe
		@param {Function} call
		@param {Object} observer
		@return {boolean} Success is true.
	*/
	unsubscribe: function(call, observer) {
		if (!call)
			return false;

		var observer = observer || this;

		// spin through the array of observers for this event, and remove one
		for (var i = 0, l = this.subscriptions.length; i < l; i++) {
			if (this.subscriptions[i].call === call && this.subscriptions[i].observer === observer)
				this.subscriptions.splice(i, 1);
		}
	},

	/**
		Called by the subject to notify observers of events.

		@method fire
		@param {any} data Optional data passed by the subject.
	*/
	fire: function(data) {
		if (typeof data == 'undefined')
			data = "";
			
		var l = this.subscriptions.length;
		
		for (var i = 0; i < l; i++) {
			this.subscriptions[i].call.call(
				this.subscriptions[i].observer,
				data,
				this.subject,
				this.subscriptions[i].data
			);
		}
	}
};

/**
	DOM manipulation classes.
	@module Dom
*/

/**
	DOM manipulation methods.
	@namespace EPOC
	@class Dom
	@static
*/
EPOC.Dom = {
	/**
		Wrapper for document.getElementById()
		@method get
		@param {string} id
		@return {DOM Element}
	*/
	get: function(id) {
		if (typeof id == "string")
			return document.getElementById(id);
		else
			return id;
	},
	detach: function(id) {
		var target = EPOC.Dom.get(id);
		target.parentNode.removeChild(target);
	},

	/**
		Get interior height of the screen in pixels
		@method getClientHeight
	*/
	getClientHeight: function() {
		return document.documentElement.clientHeight;
	},
	
	/**
		Searches for the approproate child node of the ancestor that is a parent
		of the test node (DOM Elements).
		@method getParentWithin
		@param {DOM Element} node
		@param {DOM Element} ancestor
	*/
	getParentWithin: function(node, ancestor) {
		while (node != ancestor && node.parentNode != window && node.parentNode != ancestor) {
			node = node.parentNode;
		}
		
		return node;
	},

	/**
		Check if a point is with-in a views display area
		
		@module pointWithIn
		@param {View | DOM Element} View
		@param {int} x
		@param {int} y
		@returns {boolan} true if x and y are within the views display area
	 */
	pointWithIn: function(view, x, y){
//		console.log("pointWithIn x: " + x + " y: " + y);
		if(view.container)
			view = view.container;
		if(x == undefined || y == undefined)
			return false;
		
//		console.log("left: " + view.offsetLeft + " width: " + view.clientWidth + " top: " + view.offsetTop + " height: " + view.clientHeight);
		
		if(x < view.offsetLeft || x > view.offsetLeft + view.clientWidth)
			return false;
		if(y < view.offsetTop || y > view.offsetTop + view.clientHeight)
			return false;
//		console.log("Point is in");
		return true;
	},
	
	/**
		Adds a CSS class name to a given element, leaves any others
		in the className property which might be there alone. Does not
		add the class name if it's already in the className property.
		@method addCSSClass
		@param {DOM Element} node
		@param {string} classname
	*/
	addCSSClass: function(node, classname) {
		var n = node.className.split(/\s+/);
		
		for (var i = 0, l = n.length; i < l; i++) {
			if (n[i] == classname)
				return;
		}
		n.push(classname);
		node.className = n.join(" ");
	},

	/**
		Removes a CSS class name from an element's className property
		if it exists, and leaves any other class names in the property
		intact and in order. Also called by toggle to economize on
		code.
		
		@method removeCSSClass
		@param {DOM Element} node
		@param {string} classname
	*/
	removeCSSClass: function(node, classname, toggle) {
		var n = node.className.split(/\s+/);
		
		for (var i = 0, l = n.length; i < l; i++) {
			if (n[i] == classname) {
				if (l == 1)
					node.className = "";
				else {
					n.splice(i, 1);
					node.className = n.join(" ");
				}
				return;
			}
		}

		if (toggle) {
			n.push(classname);
			node.className = n.join(" ");
		}
	},
	
	/**
		Toggles a class name from an element's className property,
		basically a wrapper for removeCSSClass.
		
		@method toggleCSSClass
		@param {DOM Element} node
		@param {string} classname
	*/
	toggleCSSClass: function(node, classname) {
		EPOC.Dom.removeCSSClass(node, classname, true);
	},
	
	/**
		Convenience method to create a new DOM element.
		TODO: Look to revamp or remove, this doesn't buy us much.
		
		@method createElement
		@param {string} tagname
		@param {string} classes
		@param {string} content
	*/
	createElement: function(tagname, classname, content) {
		var elem = document.createElement(tagname || "div");

		if (classname) {
			if (classname instanceof Array)
				var classname = classname.join(" ");
			
			elem.classname = classname;
		}

		if (content) {
			if (content instanceof Array) {
				if (typeof content[0] == 'object') {
					for (var i = 0; i < i.length; i++)
						elem.appendChild(content);
				}
				else {
					elem.innerHTML = content.join("");					
				}
			}
			else {
				if (typeof c == 'object')
					elem.appendChild(content);
				else
					elem.innerHTML = content;
			}
		}

		return elem;
	}
};

/**
	User interface classes, based on a loosly-coupled observer pattern.
	
	@namespace EPOC
	@module UI
*/
/**
	Base class for views.
	
	@namespace EPOC.UI
	@class View
	@constructor
	@param {DOM Element} container
	@param {Object} data (Optional)
*/
EPOC.UI.View = function(container, data) {
	this.data = (typeof data == 'undefined') ? [] : data;

	/**
		Fired when the data changes for any reason.
		@event changeEvent
	*/
	this.changeEvent = new EPOC.Event.Subject(this);

	/**
		Fired when the user clicks on the view.
		@event changeEvent
	*/
	this.selectEvent = new EPOC.Event.Subject(this);

	if (typeof container != 'undefined')
		this.setContainer(container);

	if (typeof data != 'undefined')
		this.setData(data);
};
EPOC.UI.View.prototype = {
	/**
		Returns the current DOM element container for our list.
		@method getContainer
		@return {DOM Element} container
	*/
	getContainer: function() {
		return this.container;
	},

	/**
		Select a DOM element for out list to display inside.
		@method setContainer
		@param {DOM Element | string} A DOM element or its ID
	*/
	setContainer: function(container) {
		this.container = EPOC.Dom.get(container);

		if (!this.container) {
			// if no container, we make one TODO: make this happen, then make sublcasses
			// build specific DOM element types (ul for listview, div for stack, etc etc).
			this.container = this.createContainer();
		}

		if(this.container){
			// set the js object so we can access it from a dom node
			this.container.jsObject = this;
		}

		this.setEvents();
	},

	/**
		Create a DOM element to contain this view. This method can be overriden
		to add classNames, ids, or any other setup ncessary to make the DOM
		part of our view.
		@method createContainer
		@param {string} classname
	*/
	createContainer: function(classname) {
		var o = document.createElement('div');

		if (classname)
			o.classname = classname;

		return o;
	},

	/**
	 	@method setEvents
	 */
	setEvents: function() {
		if (this.container.tagName == "SELECT")
			EPOC.Event.on(this.container, EPOC.Constants.selectDomEvent, this.onClick, this);	
		else
			EPOC.Event.on(this.container, EPOC.Constants.touchDomEvent, this.onClick, this);
	},

	/**
		DOM event handler, used to filter our DOM events into our custom events,
		and also update our selected list index.
		@method onClick
		@param {DOM Event} e
	*/
	onClick: function(e) {
		this.selectEvent.fire();
	},

	/**
		Free up our data and empty the DOM element container
		@method clear
	*/
	clear: function() {
		this.data = [];
		this.container.innerHTML = "";

		this.changeEvent.fire();
	},

	/**
		Renders the new data.
		@method setData
		@param {Object} data
	*/
	setData: function(data) {
		this.data = data;
		this.refresh();
	},

	/**
		Return a reference to this view's data.
		@method getData
		@return {Object} data
	*/
	getData: function() {
		return this.data;
	},

	/**
		Render our data into the DOM element container.
		@method refresh
	*/
	refresh: function() {
		if (!this.container || !this.data)
			return 0;

//		this.container.innerHTML = '';
		this.draw();

		// fire off "changed" event
		this.changeEvent.fire(this);
	},

	/**
		Re-activate this view - called when a view is going to be attached to the dom
		gives the view a chance to do anything it needs to when it is going to be viewed.
		Over ride to do something specific for your view.
		Known callers - StackView's push, pop, next, back, and home
		
		@method onShow
	 */
	onShow: function(){
	},
	
	/**
		Draw the View. Decorator method, should be augmented in subclasses or instances.
		@method draw
	*/
	draw: function() {
		this.container.innerHTML = this.data;
	},

	/**
		Clone contents from another listbox, a performance trick.
		@method addFrom
		@param {DOM Element | ListView | string} The source, either a DOM element, its ID, or another ListView
	*/
	addFrom: function(source) {
		if (typeof source == 'string')
			var source = document.getElementById(source);

		if (typeof source == 'object') {
			if (source.innerHTML)
				this.container.innerHTML = source.innerHTML;
			else if (source.container)
				this.container.innerHTML = source.container.innerHTML;

			return this.container.childNodes.length;
		}
	},

	/**
		Attach this view to a DataSource. When the DataSource changes, it will call <pre>setData()</pre> on 
		this view.
		
		@method setDataSource
		@param {DataSource} DataSource to attach.
	*/
	setDataSource: function(source) {
		this.dataSource = source;

		// need to subscribe to the changeEvent before calling setData() to avoid missing the notification
		source.loadEvent.subscribe(this.setData, this, true);
		this.setData(source.getData());
	},
	
	/**
		focusManager related methods
		
		Handle a five way key event
		
		over ride to do something
		
		@method handleFiveWay
		@param {number} key - one of EPOC.UI.FocusManager key values
		@return {boolean} true if handled
	 */
	handleFiveWay: function(key){
		return false;
	},
	
	/**
		set intial focus - called on a view after it is set as focus view in FocusManager
		
		over ride to do something
		
		@method setInitialFocus
	 */
	setInitialFocus: function(){
	}
};
